@@ -18,13 +18,16 @@
 #include <glib.h>
 #include <gmp.h>
 #include <libguile.h>
+#include <stddef.h>
+#include <sys/time.h>
 
 #include "../app.h"
+#include "../app-task.h"
 #include "../auth_int.h"
+#include "../durable-task.h"
 #include "../gzochid-auth.h"
 #include "../scheme.h"
 #include "../scheme-task.h"
-#include "../task.h"
 
 #include "task.h"
 #include "util.h"
@@ -36,7 +39,7 @@ SCM_DEFINE (primitive_schedule_task, "primitive-schedule-task", 1, 2, 0,
   gzochid_application_context *context = 
     gzochid_api_ensure_current_application_context (); 
   gzochid_auth_identity *identity =
-    gzochid_auth_identity_clone (gzochid_get_current_identity ());
+    gzochid_auth_identity_ref (gzochid_get_current_identity());
   gzochid_application_task *scheme_task = gzochid_application_task_new 
     (context, identity, gzochid_scheme_application_task_worker, callback);
 
