@@ -15,6 +15,7 @@
  * along with this program.  If not, see <http://www.gnu.org/licenses/>.
  */
 
+#include <assert.h>
 #include <glib.h>
 #include <libguile.h>
 #include <stddef.h>
@@ -51,6 +52,9 @@ gzochid_application_context_free (gzochid_application_context *app_context)
   gzochid_context *context = (gzochid_context *) app_context;
   gzochid_context_free (context);
 
+  g_hash_table_destroy (app_context->oids_to_clients);
+  g_hash_table_destroy (app_context->clients_to_oids);
+  
   g_list_free (app_context->free_oid_blocks);
 
   g_mutex_clear (&app_context->free_oids_lock);
@@ -59,7 +63,6 @@ gzochid_application_context_free (gzochid_application_context *app_context)
   gzochid_application_event_source_free (app_context->event_source);
   free (app_context->stats);
 
-  free (context->fsm->name);
   free (context);
 }
 
@@ -80,6 +83,8 @@ gzochid_with_application_context (gzochid_application_context *context,
     {
       g_private_set (&thread_application_context_key, context);
       g_private_set (&thread_identity_key, identity);
+
+      assert (context->deployment_root != NULL);
       
       scm_fluid_set_x
 	(application_root_fluid, 
