@@ -31,38 +31,64 @@
 
 #define PLUGIN_INFO_FUNCTION "gzochid_auth_init_plugin"
 
-GQuark gzochid_auth_plugin_error_quark (void)
+#define GZOCHID_AUTH_IDENTITY_CACHE_DEFAULT_MAX_SIZE 1024
+
+struct _gzochid_auth_identity
+{
+  char *name;
+  guint ref_count;
+
+  /* Indicates whether or not this identity is reserved. (I.e., is this the 
+     system identity?) */
+  
+  gboolean reserved; 
+};
+
+static gzochid_auth_identity system_identity = { "[SYSTEM]", 1, TRUE };
+
+GQuark
+gzochid_auth_plugin_error_quark (void)
 {
   return g_quark_from_static_string ("gzochid-auth-plugin-error-quark");
 }
 
-gzochid_auth_identity *gzochid_auth_identity_new (char *name)
+gzochid_auth_identity *
+gzochid_auth_identity_new (char *name)
 {
   gzochid_auth_identity *identity = calloc (1, sizeof (gzochid_auth_identity));
 
-  identity->name = name;
+  identity->name = strdup (name);
+  identity->ref_count = 1;
 
   return identity;
 }
 
-void gzochid_auth_identity_free (gzochid_auth_identity *identity)
+char *
+gzochid_auth_identity_name (gzochid_auth_identity *identity)
 {
-  free (identity->name);
-  free (identity);
+  return identity->name;
 }
 
-gzochid_auth_identity *gzochid_auth_function_pass_thru 
-(unsigned char *cred, short cred_len, gpointer auth_data, GError **error)
+void
+gzochid_auth_identity_free (gzochid_auth_identity *identity)
 {
+}
+
+gzochid_auth_identity *
+gzochid_auth_function_pass_thru (unsigned char *cred, short cred_len,
+				 gpointer auth_data, GError **error)
+{
+  char *name = NULL;
+  gzochid_auth_identity *identity = NULL;
+  
   if (cred_len <= 0)
     return NULL;
-  return gzochid_auth_identity_new (strndup ((char *) cred, cred_len));
-}
 
-gzochid_auth_identity *gzochid_auth_identity_clone
-(gzochid_auth_identity *identity)
-{
-  return gzochid_auth_identity_new (strdup (identity->name));
+  name = strndup ((char *) cred, cred_len);
+  identity = gzochid_auth_identity_new (name);
+  free (name);
+
+  return identity;
 }
 
 void 
@@ -78,22 +104,92 @@ gzochid_auth_identity_deserializer
 (gzochid_application_context *context, GString *in, GError **err)
 {
   char *name = strndup (in->str, in->len);
-  gzochid_auth_identity *identity = gzochid_auth_identity_new (name);
-
+  gzochid_auth_identity *identity =
+    gzochid_auth_identity_from_name (context->identity_cache, name);
+  
   g_string_erase (in, 0, strlen (name) + 1);
+
+  free (name);
+
   return identity;
 }
 
-void gzochid_auth_identity_finalizer
-(gzochid_application_context *context, void *ptr)
+void
+gzochid_auth_identity_finalizer (gzochid_application_context *context,
+				 void *ptr)
 {
-  gzochid_auth_identity *identity = (gzochid_auth_identity *) ptr;
-
-  free (identity->name);
-  free (identity);
+  gzochid_auth_identity_unref (ptr);
 }
 
-static gchar *remove_extension (const gchar *filename)
+struct _gzochid_auth_identity_cache
+{
+  gzochid_lru_cache *cache;
+};
+
+static gpointer
+identity_new_func (gpointer key, gpointer *key_copy)
+{
+  *key_copy = strdup (key);
+  return gzochid_auth_identity_new (key);
+}
+
+gzochid_auth_identity *
+gzochid_auth_system_identity ()
+{
+  return &system_identity;
+}
+
+gzochid_auth_identity_cache *
+gzochid_auth_identity_cache_new ()
+{
+  gzochid_auth_identity_cache *cache;
+
+  cache = malloc (sizeof (gzochid_auth_identity_cache));
+  cache->cache = gzochid_lru_cache_new_full
+    (g_str_hash, g_str_equal, identity_new_func,
+     GZOCHID_AUTH_IDENTITY_CACHE_DEFAULT_MAX_SIZE, (GDestroyNotify) free,
+     (GDestroyNotify) gzochid_auth_identity_unref);
+  
+  return cache;
+}
+
+void
+gzochid_auth_identity_cache_destroy (gzochid_auth_identity_cache *cache)
+{
+  gzochid_lru_cache_destroy (cache->cache);
+  free (cache);
+}
+
+gzochid_auth_identity *
+gzochid_auth_identity_from_name (gzochid_auth_identity_cache *cache, char *name)
+{
+  return gzochid_auth_identity_ref
+    (gzochid_lru_cache_lookup (cache->cache, name));
+}
+
+gzochid_auth_identity *
+gzochid_auth_identity_ref (gzochid_auth_identity *identity)
+{
+  if (!identity->reserved)  
+    g_atomic_int_inc (&identity->ref_count);
+
+  return identity;
+}
+
+void
+gzochid_auth_identity_unref (gzochid_auth_identity *identity)
+{
+  if (identity->reserved)
+    return;
+  
+  if (g_atomic_int_dec_and_test (&identity->ref_count)) {
+    free (identity->name);
+    free (identity);
+  }
+}
+
+static gchar *
+remove_extension (const gchar *filename)
 {
   char *dot = rindex (filename, '.');
 
@@ -102,7 +198,8 @@ static gchar *remove_extension (const gchar *filename)
   else return g_strndup (filename, dot - filename);
 }
 
-static void probe_auth_plugin (gpointer data, gpointer user_data)
+static void
+probe_auth_plugin (gpointer data, gpointer user_data)
 {
   const gchar *path = (const gchar *) data;
   gzochid_game_context *context = (gzochid_game_context *) user_data;
@@ -140,8 +237,8 @@ static void probe_auth_plugin (gpointer data, gpointer user_data)
     }
 }
 
-static void probe_auth_plugins 
-(gzochid_game_context *context, const char *search_path)
+static void
+probe_auth_plugins (gzochid_game_context *context, const char *search_path)
 {
   GDir *plugin_dir = g_dir_open (search_path, 0, NULL);
   const gchar *file = NULL;
@@ -176,14 +273,18 @@ static void probe_auth_plugins
 	      == NULL)
 	    g_sequence_insert_sorted 
 	      (plugin_queue, path, gzochid_util_string_data_compare, NULL);
+	  else g_free (path);
 	}
       
       g_sequence_foreach (plugin_queue, probe_auth_plugin, context);
       g_sequence_free (plugin_queue);
+
+      g_dir_close (plugin_dir);
     }
 }
 
-void gzochid_auth_init (gzochid_game_context *context)
+void
+gzochid_auth_init (gzochid_game_context *context)
 {
   if (g_module_supported ())
     probe_auth_plugins (context, context->auth_plugin_dir);
