@@ -34,6 +34,24 @@
 
 #define NEXT_OID_KEY 0x00
 
+struct _gzochid_data_transaction_context
+{
+  gzochid_application_context *context;
+
+  gzochid_data_oid_block *free_oids;
+  GList *used_oid_blocks;
+
+  gzochid_storage_transaction *transaction;
+
+  GHashTable *oids_to_references;
+  GHashTable *ptrs_to_references;
+
+  gboolean needs_unlock;
+};
+
+typedef struct _gzochid_data_transaction_context
+gzochid_data_transaction_context;
+
 GQuark 
 gzochid_data_error_quark (void)
 {
@@ -110,9 +128,6 @@ static gboolean
 flush_reference (gzochid_data_managed_reference *reference,
 		 gzochid_data_transaction_context *context)
 {
-  gzochid_application_data_event *event = NULL;
-  gzochid_application_event *base_event = NULL;
-
   char *oid_str = NULL;
   GString *out = NULL;
   GError *err = NULL;
@@ -148,15 +163,9 @@ flush_reference (gzochid_data_managed_reference *reference,
 	 strlen (oid_str) + 1, out->str, out->len);
       free (oid_str);
 
-      event = malloc (sizeof (gzochid_application_data_event));
-      base_event = (gzochid_application_event *) event;
-
-      base_event->type = BYTES_WRITTEN;
-      gettimeofday (&base_event->timestamp, NULL);
-      event->bytes = out->len;
-
       gzochid_application_event_dispatch
-	(context->context->event_source, base_event);
+	(context->context->event_source,
+	 gzochid_application_data_event_new (BYTES_WRITTEN, out->len));
 
       g_string_free (out, TRUE);
 
@@ -183,6 +192,8 @@ data_prepare (gpointer data)
   iface->lock (context->context->oids);
   iface->lock (context->context->names);
 
+  context->needs_unlock = TRUE;
+  
   while (reference_ptr != NULL)
     {
       if (!flush_reference (reference_ptr->data, context))
@@ -270,9 +281,12 @@ data_rollback (gpointer data)
 
       finalize_references (context);
 
-      iface->unlock (context->context->names);
-      iface->unlock (context->context->oids);
-
+      if (context->needs_unlock)
+	{
+	  iface->unlock (context->context->names);
+	  iface->unlock (context->context->oids);
+	}
+      
       transaction_context_free (context);
     }
 }
@@ -536,17 +550,10 @@ dereference (gzochid_data_transaction_context *context,
   else 
     {
       GError *local_err = NULL;
-      gzochid_application_data_event *event = 
-	malloc (sizeof (gzochid_application_data_event));
-      gzochid_application_event *base_event = 
-	(gzochid_application_event *) event;
-
-      base_event->type = BYTES_READ;
-      gettimeofday (&base_event->timestamp, NULL);
-      event->bytes = data_len;
 
       gzochid_application_event_dispatch
-	(context->context->event_source, base_event);
+	(context->context->event_source,
+	 gzochid_application_data_event_new (BYTES_READ, data_len));
 
       in = g_string_new_len (data, data_len);
       free (data);
