@@ -39,7 +39,6 @@
 #include "log.h"
 #include "httpd.h"
 #include "stats.h"
-#include "threads.h"
 
 #define OID_PREFIX_LEN 26
 #define OID_SUFFIX_LEN 29
@@ -420,6 +419,17 @@ app_info (struct MHD_Connection *connection,
 			  app_context->stats->bytes_written);
   g_string_append (response_str, "      </tr>\n");
 
+  g_string_append (response_str, "      <tr>\n");
+  g_string_append (response_str, "        <td>Messages received</td>\n");
+  g_string_append_printf (response_str, "        <td>%u</td>\n", 
+			  app_context->stats->num_messages_received);
+  g_string_append (response_str, "      </tr>\n");
+  g_string_append (response_str, "      <tr>\n");
+  g_string_append (response_str, "        <td>Messages sent</td>\n");
+  g_string_append_printf (response_str, "        <td>%u</td>\n", 
+			  app_context->stats->num_messages_sent);
+  g_string_append (response_str, "      </tr>\n");
+
   g_string_append (response_str, "      <tr>\n");
   g_string_append (response_str, "        <td>Transactions started</td>\n");
   g_string_append_printf (response_str, "        <td>%u</td>\n", 
@@ -561,29 +571,19 @@ dispatch (void *cls, struct MHD_Connection *connection,
   return ret;
 }
 
-static void 
-initialize_async (gpointer data, gpointer user_data)
-{
-  gzochid_context *context = data;
-  gzochid_fsm_to_state (context->fsm, GZOCHID_HTTPD_STATE_RUNNING);
-}
-
 static void 
 initialize (int from_state, int to_state, gpointer user_data)
 {
   gzochid_context *context = user_data;
   gzochid_httpd_context *httpd_context = (gzochid_httpd_context *) context;
-  gzochid_admin_context *admin_context = 
-    (gzochid_admin_context *) context->parent;
 
   httpd_context->daemon = MHD_start_daemon 
     (MHD_USE_SELECT_INTERNALLY, httpd_context->port, NULL, NULL, dispatch, 
      httpd_context, MHD_OPTION_END);
 
   gzochid_notice ("HTTP server listening on port %d", httpd_context->port);
-
-  gzochid_thread_pool_push 
-    (admin_context->pool, initialize_async, context, NULL);
+  
+  gzochid_fsm_to_state (context->fsm, GZOCHID_HTTPD_STATE_RUNNING);
 }
 
 gzochid_httpd_context *
