@@ -26,6 +26,7 @@
 #include "app-task.h"
 #include "callback.h"
 #include "data.h"
+#include "durable-task.h"
 #include "guile.h"
 #include "gzochid-auth.h"
 #include "io.h"
@@ -164,12 +165,7 @@ scm_to_callback (gzochid_application_context *context, SCM scm_callback)
     gzochid_data_create_reference 
     (context, &gzochid_scheme_data_serialization, scm_callback);
 
-  mpz_t oid;
-  
-  mpz_init (oid);
-  mpz_set (oid, reference->oid);      
-    
-  return gzochid_application_callback_new (procedure, module, oid);
+  return gzochid_application_callback_new (procedure, module, reference->oid);
 }
 
 static gpointer 
@@ -406,10 +402,13 @@ gzochid_scheme_application_disconnected_worker
 
   if (session->handler == NULL)
     {
+      /* In that case, don't bother trying to dredge up its disconnect handler.
+	 Just jump straight to removing the session from the data store. */
+      
       gzochid_tx_warning 
 	(context, "Session '%s' disconnected after incomplete login.", ptr);
-
-      gzochid_client_session_disconnected_worker (context, identity, oid_str);
+      gzochid_client_session_disconnected_worker (context, identity, ptr);
+      
       return;
     }
 
@@ -431,7 +430,6 @@ gzochid_scheme_application_disconnected_worker
 	      oid_str);
 
       g_error_free (err);      
-      gzochid_client_session_disconnected_worker (context, identity, oid_str);
       return;
     }
 
@@ -451,8 +449,15 @@ gzochid_scheme_application_disconnected_worker
 	     (scm_variable_ref (exception_var)));
 	}
     }
+  else gzochid_client_session_disconnected_worker (context, identity, ptr);
+}
 
-  gzochid_client_session_disconnected_worker (context, identity, oid_str);
+void 
+gzochid_scheme_application_disconnected_cleanup_worker 
+(gzochid_application_context *context, gzochid_auth_identity *identity, 
+ gpointer ptr)
+{
+  free (ptr);
 }
 
 void
@@ -535,7 +540,8 @@ gzochid_scheme_task_new (gzochid_application_context *context,
 static void 
 bind_scm (char *module, SCM *binding, char *name)
 {
-  SCM var = scm_c_public_variable (module, name);
+  SCM var = scm_module_variable
+    (scm_c_resolve_module (module), scm_from_locale_symbol (name));
 
   if (scm_is_false (var))
     g_error ("Missing Scheme binding for `%s'. Aborting...", name);
