@@ -15,6 +15,7 @@
  * along with this program.  If not, see <http://www.gnu.org/licenses/>.
  */
 
+#include <assert.h>
 #include <glib.h>
 #include <gmp.h>
 #include <gzochi-common.h>
@@ -113,9 +114,16 @@ gzochid_scheme_invoke (gzochid_application_context *context,
 static SCM 
 resolve_procedure (char *procedure, GList *module)
 {
-  SCM scm_procedure = scm_from_locale_symbol (procedure);
-  SCM scm_module = gzochid_scheme_glist_to_list 
+  SCM scm_procedure;
+  SCM scm_module;
+
+  assert (procedure != NULL);
+  assert (module != NULL);
+
+  scm_procedure = scm_from_locale_symbol (procedure);
+  scm_module = gzochid_scheme_glist_to_list 
     (module, (SCM (*) (gpointer)) scm_from_locale_symbol);
+  
   return scm_public_ref (scm_module, scm_procedure);
 }
 
@@ -187,10 +195,18 @@ scheme_managed_record_deserializer (gzochid_application_context *context,
   SCM exception_var = scm_make_variable (SCM_UNSPECIFIED);
   GList *args = g_list_append 
     (g_list_append (g_list_append (NULL, "gzochi"), "private"), "data");
-     
-  g_string_erase (in, 0, 4);
-  vec = scm_take_u8vector ((unsigned char *) in->str, vec_len);
 
+  g_string_erase (in, 0, 4);
+
+  if (vec_len > in->len)
+    {
+      g_set_error
+	(err, GZOCHID_SCHEME_ERROR, GZOCHID_SCHEME_ERROR_SERIAL,
+	 "Incorrect length prefix for managed record; possibly corrupt store.");
+      return SCM_BOOL_F;
+    }
+  
+  vec = scm_take_u8vector ((unsigned char *) in->str, vec_len);
   port = scm_open_bytevector_input_port (vec, SCM_BOOL_F);
 
   record = gzochid_scheme_invoke_callback
@@ -489,9 +505,10 @@ gzochid_scheme_create_client_session (gzochid_client_session *session,
   char *oid_str = mpz_get_str (NULL, 16, oid);
 
   SCM scm_oid = scm_c_locale_stringn_to_number (oid_str, strlen (oid_str), 16);
-  SCM ret = scm_call_2 (scm_make_client_session, 
-			scm_from_locale_string (session->identity->name),
-			scm_oid);
+  SCM ret = scm_call_2
+    (scm_make_client_session, 
+     scm_from_locale_string (gzochid_auth_identity_name (session->identity)),
+     scm_oid);
 
   scm_gc_protect_object (ret);
   free (oid_str);
@@ -566,7 +583,8 @@ gzochid_scheme_create_periodic_task_handle (mpz_t oid)
 static void 
 bind_scm (char *module, SCM *binding, char *name)
 {
-  SCM var = scm_c_public_variable (module, name);
+  SCM var = scm_module_variable
+    (scm_c_resolve_module (module), scm_from_locale_symbol (name));
 
   if (scm_is_false (var))
     g_error ("Missing Scheme binding for `%s'. Aborting...", name);
