@@ -25,6 +25,7 @@
 #include "app.h"
 #include "auth_int.h"
 #include "data.h"
+#include "event.h"
 #include "game.h"
 #include "gzochid-auth.h"
 #include "io.h"
@@ -81,6 +82,12 @@ struct _gzochid_client_session_transaction_context
 typedef struct _gzochid_client_session_transaction_context
 gzochid_client_session_transaction_context;
 
+GQuark
+gzochid_session_error_quark ()
+{
+  return g_quark_from_static_string ("gzochid-session-error-quark");
+}
+
 static gzochid_client_session_transaction_context *
 create_transaction_context (gzochid_application_context *context)
 {
@@ -168,6 +175,9 @@ static void
 cleanup_transaction (gzochid_client_session_transaction_context *tx_context)
 {
   g_list_free_full (tx_context->operations, free_operation);
+
+  if (tx_context->login_operation != NULL)
+    free_operation (tx_context->login_operation);
   
   free (tx_context);
 }
@@ -212,6 +222,9 @@ session_commit_operation
 		      tx_context->login_operation->target_session) != 0)
 	{
 	  msg_op = (gzochid_client_session_pending_message_operation *) op;
+	  gzochid_application_event_dispatch
+	    (context->event_source,
+	     gzochid_application_event_new (MESSAGE_SENT));
 	  gzochid_protocol_client_send (client, msg_op->message, msg_op->len);
 	}
 
@@ -288,6 +301,8 @@ deserialize_client_session (gzochid_application_context *context, GString *in,
   if (gzochid_util_deserialize_boolean (in))
     session->handler = deserialize_handler (context, in, NULL);
 
+  gzochid_auth_identity_unref (identity);
+  
   return session;
 }
 
@@ -399,16 +414,22 @@ remove_session (gzochid_application_context *context, const char *oid_str,
 		gzochid_info 
 		  ("Unable to remove Scheme object for session '%s': %s",
 		   oid_str, local_err->message);
-
-	      g_clear_error (&local_err);
 	    }
-	    
-	    gzochid_data_remove_object (session_reference, &local_err);
-	  
-	  if (local_err != NULL)
-	    g_propagate_error (err, local_err);	  
-	}
 
+	  /* Don't attempt to remove the object if the transaction is in a
+	     failed state - it's a waste of time and tends to irritate the 
+	     storage engine. */
+	  
+	  if (!g_error_matches
+	      (local_err, GZOCHID_DATA_ERROR, GZOCHID_DATA_ERROR_TRANSACTION))
+	    {
+	      g_clear_error (&local_err);
+	      gzochid_data_remove_object (session_reference, &local_err);
+	    }
+	}
+      if (local_err != NULL)
+	g_propagate_error (err, local_err);	  
+	
       mpz_clear (oid);
     }
   else g_propagate_error (err, local_err);
@@ -422,7 +443,6 @@ gzochid_client_session_disconnected_worker
  gpointer data)
 {
   remove_session (context, data, NULL);
-  free (data);
 }
 
 gzochid_client_session *
@@ -431,7 +451,7 @@ gzochid_client_session_new (gzochid_auth_identity *identity)
   gzochid_client_session *session = (gzochid_client_session *)
     calloc (1, sizeof (gzochid_client_session));
 
-  session->identity = gzochid_auth_identity_clone (identity);
+  session->identity = gzochid_auth_identity_ref (identity);
   session->channels = g_sequence_new (NULL);
   mpz_init (session->scm_oid);
 
@@ -441,8 +461,10 @@ gzochid_client_session_new (gzochid_auth_identity *identity)
 void 
 gzochid_client_session_free (gzochid_client_session *session)
 {
-  mpz_clear (session->scm_oid);
+  gzochid_auth_identity_unref (session->identity);
   g_sequence_free (session->channels);
+  mpz_clear (session->scm_oid);
+  
   free (session);
 }
 
@@ -580,12 +602,6 @@ gzochid_prefix_binding_persistence_task_data_new
   return task;
 }
 
-void 
-gzochid_persistence_task_data_free (gzochid_persistence_task_data *task)
-{
-  free (task);
-}
-
 static void 
 persistence_task_worker (gzochid_application_context *context, 
 			 gzochid_auth_identity *identity, gpointer data)
@@ -599,9 +615,8 @@ persistence_task_worker (gzochid_application_context *context,
 }
 
 static void 
-prefix_binding_persistence_task_worker (gzochid_application_context *context,
-					gzochid_auth_identity *identity, 
-					gpointer data)
+prefix_task_worker (gzochid_application_context *context,
+		    gzochid_auth_identity *identity, gpointer data)
 {
   gzochid_prefix_binding_persistence_task_data *
     prefix_binding_persistence_task = data;
@@ -622,6 +637,13 @@ prefix_binding_persistence_task_worker (gzochid_application_context *context,
   g_string_free (binding, TRUE);
 }
 
+static void
+prefix_task_cleanup_worker (gzochid_application_context *context,
+			    gzochid_auth_identity *identity, gpointer data)
+{
+  free (data);
+}
+
 gzochid_task *
 gzochid_data_prefix_binding_persistence_task_new
 (gzochid_application_context *context, gzochid_auth_identity *identity, 
@@ -631,36 +653,34 @@ gzochid_data_prefix_binding_persistence_task_new
   gzochid_prefix_binding_persistence_task_data *task_data =
     gzochid_prefix_binding_persistence_task_data_new 
     (serialization, data, holder, prefix);
-  gzochid_application_task *transactional_task = 
-    gzochid_application_task_new 
-    (context, identity, prefix_binding_persistence_task_worker, task_data);
+
+  gzochid_application_task *task = gzochid_application_task_new
+    (context, identity, prefix_task_worker, task_data);
+  gzochid_application_task *cleanup_task = gzochid_application_task_new 
+    (context, identity, prefix_task_cleanup_worker, task_data);
   gzochid_transactional_application_task_execution *execution = 
-    gzochid_transactional_application_task_execution_new 
-    (transactional_task, NULL);
+    gzochid_transactional_application_task_execution_new
+    (task, NULL, cleanup_task);
+
+  /* Not necessary to hold a ref to these, as we've transferred them to the
+     execution. */
+  
+  gzochid_application_task_unref (task);
+  gzochid_application_task_unref (cleanup_task);
+  
+  /* This task uses the re-executing task worker (in concert with 
+     `gzochid_schedule_run_task' below) to ensure that the full transactional
+     execution lifecycle - including retries - is completed before the oid 
+     holder is freed. */
+
   gzochid_application_task *application_task = gzochid_application_task_new 
-    (context, identity, 
-     gzochid_application_resubmitting_transactional_task_worker, execution);
+    (context, identity,
+     gzochid_application_reexecuting_transactional_task_worker, execution);
 
   return gzochid_task_immediate_new 
     (gzochid_application_task_thread_worker, application_task);
 }
 
-void 
-gzochid_data_prefix_binding_persistence_task_free (gzochid_task *task)
-{
-  gzochid_application_task *application_task = task->data;
-  gzochid_transactional_application_task_execution *execution = 
-    application_task->data;
-  gzochid_application_task *transactional_task = execution->task;
-  gzochid_persistence_task_data *data = transactional_task->data;
-  
-  free (data);
-  free (transactional_task);
-  gzochid_transactional_application_task_execution_free (execution);
-  free (application_task);
-  free (task);
-}
-
 void 
 gzochid_client_session_persist (gzochid_application_context *context, 
 				gzochid_client_session *session, mpz_t oid)
@@ -673,28 +693,27 @@ gzochid_client_session_persist (gzochid_application_context *context,
      session, holder, SESSION_PREFIX);
 
   gzochid_schedule_run_task (game_context->task_queue, task);  
+  gzochid_task_free (task);
 
   mpz_set (oid, holder->oid);
-
   gzochid_oid_holder_free (holder);
-  gzochid_data_prefix_binding_persistence_task_free (task);
 }
 
 void 
-gzochid_sweep_client_sessions (gzochid_application_context *context, 
-			       gzochid_auth_identity *identity)
+gzochid_sweep_client_sessions (gzochid_application_context *context,
+			       GError **err)
 {
   mpz_t oid;
-  GError *err = NULL;
+  GError *local_err = NULL;
   char *next_binding = NULL;
   int prefix_len = strlen (SESSION_PREFIX), num_sessions = 0;
 
   mpz_init (oid);
   gzochid_tx_info (context, "Sweeping old client sessions.");
   next_binding = gzochid_data_next_binding_oid 
-    (context, SESSION_PREFIX, oid, &err);
+    (context, SESSION_PREFIX, oid, &local_err);
 
-  assert (err == NULL);
+  assert (local_err == NULL);
 
   while (next_binding != NULL 
          && strncmp (SESSION_PREFIX, next_binding, prefix_len) == 0)
@@ -703,14 +722,27 @@ gzochid_sweep_client_sessions (gzochid_application_context *context,
       char *next_next_binding = NULL;
       
       gzochid_scheme_application_disconnected_worker 
-	(context, identity, oid_str);      
+	(context, gzochid_auth_system_identity (), oid_str);      
 
-      assert (! gzochid_transaction_rollback_only ());
-      
+      if (gzochid_transaction_rollback_only ())
+	{
+	  g_set_error
+	    (err, GZOCHID_SESSION_ERROR, GZOCHID_SESSION_ERROR_DISCONNECT,
+	     "Transaction rollback while sweeping disconnected session '%s'.",
+	     oid_str);
+	  
+	  free (next_binding);
+	  free (oid_str);
+	  mpz_clear (oid);
+	  
+	  return;
+	}
+
+      free (oid_str);
       next_next_binding = gzochid_data_next_binding_oid 
-	(context, next_binding, oid, &err);
+	(context, next_binding, oid, &local_err);
 
-      assert (err == NULL);
+      assert (local_err == NULL);
 
       free (next_binding);
       next_binding = next_next_binding;
