@@ -26,6 +26,18 @@
 #include "gzochid-auth.h"
 #include "io.h"
 
+#define GZOCHID_SESSION_ERROR gzochid_session_error_quark ()
+
+GQuark gzochid_session_error_quark (void);
+
+enum _GzochidSessionError
+  {
+    GZOCHID_SESSION_ERROR_DISCONNECT,
+    GZOCHID_SESSION_ERROR_FAILED
+  };
+
+typedef enum _GzochidSessionError GzochidSessionError;
+  
 struct _gzochid_client_session_handler
 {
   gzochid_application_callback *received_message;
@@ -49,6 +61,11 @@ typedef struct _gzochid_client_session gzochid_client_session;
 
 extern gzochid_io_serialization gzochid_client_session_serialization;
 
+/* Transactional application worker that removes the persisted session 
+   corresponding to the specified oid string, as well as its binding in the
+   "names" database. The Scheme representation of the session is also removed if
+   it exists. */
+
 void gzochid_client_session_disconnected_worker
 (gzochid_application_context *, gzochid_auth_identity *, gpointer);
 
@@ -66,7 +83,6 @@ void gzochid_client_session_send_message
 
 void gzochid_client_session_persist 
 (gzochid_application_context *, gzochid_client_session *, mpz_t);
-void gzochid_sweep_client_sessions 
-(gzochid_application_context *, gzochid_auth_identity *);
+void gzochid_sweep_client_sessions (gzochid_application_context *, GError **);
 
 #endif /* GZOCHID_SESSION_H */
