@@ -1,5 +1,5 @@
 /* threads.h: Prototypes and declarations for threads.c
- * Copyright (C) 2011 Julian Graham
+ * Copyright (C) 2013 Julian Graham
  *
  * gzochi is free software: you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by
@@ -22,6 +22,12 @@
 
 typedef void (*gzochid_thread_worker) (gpointer, gpointer);
 
+typedef struct _gzochid_thread_work
+{
+  gzochid_thread_worker worker;
+  gpointer data;
+} gzochid_thread_work;
+
 GThreadPool *gzochid_thread_pool_new (gpointer, gint, gboolean, GError **);
 void gzochid_thread_pool_push 
 (GThreadPool *, gzochid_thread_worker, gpointer, GError **);
